import asyncio
import logging
from typing import Dict

from .const import (
    MAX_REQUEST_PARAMETERS,
    PARAM_HOTWATER_SYSTEMS,
    PARAM_CLIMATE_SYSTEMS,
    PARAM_COMPRESSOR_FREQUENCY,
    PARAM_PUMP_SPEED_HEATING_MEDIUM,
    PARAM_STATUS_COOLING,
    PARAM_VENTILATION_SYSTEMS,
    SMARTHOME_MODES,
)
from .typing import StatusItemIcon, SetThermostatModel, Thermostat, SmartHomeSystem
from .types import (
    VentilationSystem,
    ClimateSystem,
    HotWaterSystem,
)

from .monitor import Monitor
from .uplink import Uplink
from .session import UplinkSession

_LOGGER = logging.getLogger(__name__)


async def get_active_climate(
    uplink: Uplink, system_id: int
) -> Dict[str, ClimateSystem]:
    active = {}

    async def check(key: str, value: ClimateSystem):
        if value.active_accessory is None:
            active[key] = value
            return

        available = await uplink.get_parameter(system_id, value.active_accessory)

        _LOGGER.debug("Climate %s:%s active_accessory: %s", system_id, key, available)
        if available and available["rawValue"] == 1:
            active[key] = value

    await asyncio.gather(
        *[check(key, value) for key, value in PARAM_CLIMATE_SYSTEMS.items()]
    )

    return active


async def get_active_hotwater(
    uplink: Uplink, system_id: int
) -> Dict[str, HotWaterSystem]:
    active = {}

    async def check(key: str, value: HotWaterSystem):
        if value.hot_water_production is None:
            active[key] = value
            return

        available = await uplink.get_parameter(system_id, value.hot_water_production)

        _LOGGER.debug(
            "Hotwater %s:%s hot_water_production: %s", system_id, key, available
        )
        if available and available["rawValue"] == 1:
            active[key] = value

    await asyncio.gather(
        *[check(key, value) for key, value in PARAM_HOTWATER_SYSTEMS.items()]
    )

    return active


async def get_active_ventilations(
    uplink: Uplink, system_id: int
) -> Dict[str, VentilationSystem]:
    active = {}

    async def check(key: str, value: VentilationSystem):
        if value.fan_speed is None:
            return

        available = await uplink.get_parameter(system_id, value.fan_speed)

        _LOGGER.debug("Ventilation %s:%s fan_speed: %s", system_id, key, available)
        if available and available["rawValue"] != -32768:
            active[key] = value

    await asyncio.gather(
        *[check(key, value) for key, value in PARAM_VENTILATION_SYSTEMS.items()]
    )

    return active



from typing import Dict, NewType, Any, Union, Optional, List
from typing_extensions import (
    TypedDict
)

SystemId = int
Identifier = Union[str, int]
ParameterId = Identifier
Parameter = Dict[str, Any]
StatusItemIcon = Dict[str, Any]
ParameterSet = Dict[ParameterId, Parameter]


class SmartHomeSystem(TypedDict):
    name: str


class Thermostat(TypedDict):
    smartHomeSystem: SmartHomeSystem
    name: str
    climateSystems: Optional[List[int]]
    currentTemperature: Optional[str]
    targetTemperature: Optional[str]


class SetThermostatModel(TypedDict):
    externalId: int
    name: str
    actualTemp: Optional[int] # Multiplied by 10
    targetTemp: Optional[int] # Multiplied by 10
    valvePosition: Optional[int]
    climateSystems: Optional[List[int]]



"""Handler for uplink."""
import attr
import logging
import asyncio
import aiohttp
from datetime import datetime, timedelta
from typing import List, Optional, Any

from .utils import chunks, chunk_pop
from .typing import (
    StatusItemIcon,
    ParameterId,
    Thermostat,
    SetThermostatModel,
)
from .const import MAX_REQUEST_PARAMETERS

_LOGGER = logging.getLogger(__name__)


class ParameterRequest:
    def __init__(self, parameter_id: str):
        self.parameter_id = parameter_id
        self.data = None
        self.done = False


class Throttle:
    """
    Throttling requests to API.

    Works by awaiting our turn then executing the request,
    and scheduling next request at a delay after the previous
    request completed.
    """

    def __init__(self, delay):
        self._delay = delay
        self._timestamp = datetime.now()

    async def __aenter__(self):
        timestamp = datetime.now()
        delay = (self._timestamp - timestamp).total_seconds()
        if delay > 0:
            _LOGGER.debug("Delaying request by %s seconds due to throttle", delay)
            await asyncio.sleep(delay)
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        self._timestamp = datetime.now() + self._delay


class Uplink:
    def __init__(
        self, session, loop=None, base="https://api.nibeuplink.com", throttle=4.5
    ):

        self.state = None
        self.lock = asyncio.Lock()
        self.throttle = Throttle(timedelta(seconds=throttle))
        self.session = session
        self.loop = loop
        self.base = base
        self.requests = {}

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        pass

    async def get(self, url, *args, **kwargs):
        return await self.session.request(
            "GET", f"{self.base}/api/v1/{url}", *args, **kwargs
        )

    async def put(self, url, *args, **kwargs):
        return await self.session.request(
            "PUT", f"{self.base}/api/v1/{url}", *args, **kwargs
        )

    async def post(self, url, *args, **kwargs):
        return await self.session.request(
            "POST", f"{self.base}/api/v1/{url}", *args, **kwargs
        )

    async def get_parameter_raw(self, system_id: int, parameter_id: ParameterId):

        request = ParameterRequest(str(parameter_id))
        if system_id not in self.requests:
            self.requests[system_id] = []
        self.requests[system_id].append(request)

        # yield to any other runnable that want to add requests
        await asyncio.sleep(0)

        while True:
            async with self.lock:

                # check if we are already finished, by somebody elses request
                if request.done:
                    break

                if len(self.requests[system_id]) == 0:
                    break

                async with self.throttle:
                    # chop of as many requests from start as possible
                    requests = chunk_pop(
                        self.requests[system_id], MAX_REQUEST_PARAMETERS
                    )

                    _LOGGER.debug(
                        "Requesting parameters {}".format(
                            [str(x.parameter_id) for x in requests]
                        )
                    )

                    data = await self.get(
                        f"systems/{system_id}/parameters",
                        params=[
                            ("parameterIds", str(x.parameter_id)) for x in requests
                        ],
                        headers={},
                    )

                lookup = {p["name"]: p for p in data}

                for r in requests:
                    r.done = True
                    if r.parameter_id in lookup:
                        r.data = lookup[r.parameter_id]

        return request.data

    def add_parameter_extensions(self, data):
        if data:
            if data["displayValue"].endswith(data["unit"]) and len(data["unit"]):
                value = data["displayValue"][: -len(data["unit"])]

                try:
                    value = float(value)
                except ValueError:
                    pass

                data["value"] = value
            elif data["displayValue"] == "--":
                data["value"] = None
            else:
                data["value"] = data["displayValue"]

    async def get_parameter(self, system_id: int, parameter_id: ParameterId):
        data = await self.get_parameter_raw(system_id, parameter_id)
        self.add_parameter_extensions(data)
        return data

    async def put_parameter(
        self, system_id: int, parameter_id: ParameterId, value: Any
    ):
        headers = {
            "Accept": "application/json",
            "Content-Type": "application/json;charset=UTF-8",
        }

        data = {"settings": {str(parameter_id): value}}
        async with self.lock, self.throttle:
            result = await self.put(
                f"systems/{system_id}/parameters", json=data, headers=headers,
            )
        return result[0]["status"]

    async def get_system(self, system_id: int):
        _LOGGER.debug("Requesting system {}".format(system_id))
        async with self.lock, self.throttle:
            return await self.get(f"systems/{system_id}")

    async def get_systems(self):
        _LOGGER.debug("Requesting systems")
        async with self.lock, self.throttle:
            data = await self.get("systems")
            return data["objects"]

    async def get_category_raw(
        self, system_id: int, category_id: str, unit_id: int = 0
    ):
        _LOGGER.debug(
            "Requesting category {} on system {}".format(category_id, system_id)
        )
        async with self.lock, self.throttle:
            return await self.get(
                f"systems/{system_id}/serviceinfo/categories/{category_id}",
                {"systemUnitId": unit_id},
            )

    async def get_category(self, system_id: int, category_id: str, unit_id: int = 0):
        data = await self.get_category_raw(system_id, category_id, unit_id)
        for param in data:
            self.add_parameter_extensions(param)
        return data

    async def get_categories(self, system_id: int, parameters: bool, unit_id: int = 0):
        _LOGGER.debug("Requesting categories on system {}".format(system_id))

        async with self.lock, self.throttle:
            data = await self.get(
                f"systems/{system_id}/serviceinfo/categories",
                params={"parameters": str(parameters), "systemUnitId": unit_id},
            )
        for category in data:
            if category["parameters"]:
                for param in category["parameters"]:
                    self.add_parameter_extensions(param)
        return data

    async def get_status_raw(self, system_id: int):
        _LOGGER.debug("Requesting status on system {}".format(system_id))
        async with self.lock, self.throttle:
            return await self.get(f"systems/{system_id}/status/system")

    async def get_status(self, system_id: int) -> List[StatusItemIcon]:
        data = await self.get_status_raw(system_id)
        for status in data:
            if status["parameters"]:
                for param in status["parameters"]:
                    self.add_parameter_extensions(param)
        return data

    async def get_units(self, system_id: int):
        _LOGGER.debug("Requesting units on system {}".format(system_id))
        async with self.lock, self.throttle:
            return await self.get(f"systems/{system_id}/units")

    async def get_unit_status(self, system_id: int, unit_id: int):
        _LOGGER.debug("Requesting unit {} on system {}".format(unit_id, system_id))
        async with self.lock, self.throttle:
            data = await self.get(f"systems/{system_id}/status/systemUnit/{unit_id}")
        for status in data:
            if status["parameters"]:
                for param in status["parameters"]:
                    self.add_parameter_extensions(param)
        return data

    async def get_notifications(
        self, system_id: int, active: bool = True, notifiction_type: str = "ALARM"
    ):
        _LOGGER.debug("Requesting notifications on system {}".format(system_id))
        params = {
            "active": str(active),
            "itemsPerPage": 100,
            "type": notifiction_type,
        }
        async with self.lock, self.throttle:
            data = await self.get(f"systems/{system_id}/notifications", params=params)
        return data["objects"]

    async def get_smarthome_mode(self, system_id: int) -> str:
        async with self.lock, self.throttle:
            data = await self.get(f"systems/{system_id}/smarthome/mode")
        mode = data["mode"]
        _LOGGER.debug("Get smarthome mode %s", mode)
        return mode

    async def put_smarthome_mode(self, system_id: int, mode: str) -> None:
        headers = {
            "Accept": "application/json",
            "Content-Type": "application/json;charset=UTF-8",
        }

        data = {"mode": mode}
        async with self.lock, self.throttle:
            data = await self.session.put(
                f"systems/{system_id}/smarthome/mode", json=data, headers=headers,
            )
        _LOGGER.debug("Set smarthome mode %s -> %s", mode, data)

    async def get_smarthome_thermostats(self, system_id: int) -> List[Thermostat]:
        async with self.lock, self.throttle:
            data = await self.get(f"systems/{system_id}/smarthome/thermostats")
        _LOGGER.debug("Get smarthome thermostats %s", data)
        return data

    async def post_smarthome_thermostats(
        self, system_id: int, thermostat: SetThermostatModel
    ) -> None:
        headers = {
            "Accept": "application/json",
            "Content-Type": "application/json;charset=UTF-8",
        }

        _LOGGER.debug("Post smarthome thermostat: %s", thermostat)
        async with self.lock, self.throttle:
            await self.post(
                f"systems/{system_id}/smarthome/thermostats", json=thermostat, headers=headers,
            )



import logging
import sys
import asyncio
import json
from urllib.parse import urldefrag, parse_qs

import nibeuplink
import argparse

_LOGGER = logging.getLogger(__name__)


def pair(arg):
    data = arg.split("=")
    return (data[0], data[1])


parser = argparse.ArgumentParser(description="Read data from nibe uplink.")
parser.add_argument("--client_id", required=True)
parser.add_argument("--client_secret", required=True)
parser.add_argument("--redirect_uri", required=True)

parser.add_argument("--system", type=int)
parser.add_argument("--categories", action="store_true")
parser.add_argument("--category", nargs="+")
parser.add_argument("--status", action="store_true")
parser.add_argument("--parameter", nargs="+", type=str)
parser.add_argument("--put_parameter", nargs="+", type=pair)
parser.add_argument("--units", action="store_true")
parser.add_argument("--alarms", action="store_true")
parser.add_argument("--info", action="store_true")
parser.add_argument("--unit", type=int, default=0)
parser.add_argument("--unit_status", action="store_true")
parser.add_argument("--subsystems", action="store_true")
parser.add_argument("--verbose", action="store_true")
parser.add_argument("--smarthome_mode", action="store_true")
parser.add_argument("--put_smarthome_mode", type=str)
parser.add_argument("--smarthome_thermostats", action="store_true")
parser.add_argument("--post_smarthome_thermostats", type=dict)

args = parser.parse_args()

if args.verbose:
    root = logging.getLogger()
    root.setLevel(logging.DEBUG)

    ch = logging.StreamHandler(sys.stdout)
    ch.setLevel(logging.DEBUG)
    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )
    ch.setFormatter(formatter)
    root.addHandler(ch)


STORE = "nibeuplink.json"


def token_read():
    try:
        with open(STORE, "r") as myfile:
            return json.load(myfile)
    except FileNotFoundError:
        return None


def token_write(token):
    with open(STORE, "w") as myfile:
        json.dump(token, myfile, indent=2)


async def run():

    scope = ["READSYSTEM"]
    if args.put_parameter or args.put_smarthome_mode:
        scope.append("WRITESYSTEM")

    async with nibeuplink.UplinkSession(
        client_id=args.client_id,
        client_secret=args.client_secret,
        redirect_uri=args.redirect_uri,
        access_data=token_read(),
        access_data_write=token_write,
        scope=scope,
    ) as session:

        if not session.access_data:
            auth_uri = session.get_authorize_url()
            print(auth_uri)
            result = input("Enter full redirect url: ")
            await session.get_access_token(session.get_code_from_url(result))

        uplink = nibeuplink.Uplink(session)

        todo = []
        if not args.system:
            todo.extend([uplink.get_systems()])
        else:

            if args.parameter:
                todo.extend(
                    [uplink.get_parameter(args.system, p) for p in args.parameter]
                )

            if args.categories:
                todo.extend([uplink.get_categories(args.system, False, args.unit)])

            if args.category:
                todo.extend(
                    [
                        uplink.get_category(args.system, p, args.unit)
                        for p in args.category
                    ]
                )

            if args.status:
                todo.extend([uplink.get_status(args.system)])

            if args.unit_status:
                todo.extend([uplink.get_unit_status(args.system, args.unit)])

            if args.units:
                todo.extend([uplink.get_units(args.system)])

            if args.alarms:
                todo.extend([uplink.get_notifications(args.system)])

            if args.info:
                todo.extend([uplink.get_notifications(args.system, notifiction_type=1)])

            if args.put_parameter:
                todo.extend(
                    [
                        uplink.put_parameter(args.system, p[0], p[1])
                        for p in args.put_parameter
                    ]
                )

            if args.smarthome_mode:
                todo.extend([uplink.get_smarthome_mode(args.system)])

            if args.put_smarthome_mode:
                todo.extend(
                    [uplink.put_smarthome_mode(args.system, args.put_smarthome_mode)]
                )

            if args.smarthome_thermostats:
                todo.extend([uplink.get_smarthome_thermostats(args.system)])

            if args.post_smarthome_thermostats:
                todo.extend(
                    [
                        uplink.post_smarthome_thermostats(
                            args.system, args.post_smarthome_thermostats
                        )
                    ]
                )

            if args.subsystems:

                async def named(coroutine):
                    data = await coroutine
                    return [x.name for x in data.values()]

                todo.extend(
                    [
                        named(nibeuplink.get_active_climate(uplink, args.system)),
                        named(nibeuplink.get_active_hotwater(uplink, args.system)),
                        named(nibeuplink.get_active_ventilations(uplink, args.system)),
                    ]
                )

            if not len(todo):
                todo.extend([uplink.get_system(args.system)])

        res = await asyncio.gather(*todo)
        for a in res:
            try:
                print(json.dumps(a, indent=1))
            except TypeError:
                print(a)


def main():
    loop = asyncio.get_event_loop()
    loop.run_until_complete(run())


if __name__ == "__main__":
    main()
